from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import warnings

__all__ = ['draw_contour',
           'draw_mncontour',
           'draw_profile',
           'mncontour_grid',
           ]


def draw_profile(self, vname, x, y, s=None, band=True, text=True):
    from matplotlib import pyplot as plt
    import numpy as np

    x = np.array(x)
    y = np.array(y)
    if s is not None:
        s = np.array(s, dtype=bool)
        x = x[s]
        y = y[s]

    plt.plot(x, y)
    plt.grid(True)
    plt.xlabel(vname)
    plt.ylabel('FCN')

    try:
        minpos = np.argmin(y)
        ymin = y[minpos]
        tmpy = y - ymin
        # now scan for minpos to the right until greater than one
        up = self.errordef
        righty = np.power(tmpy[minpos:] - up, 2)
        right_min = np.argmin(righty)
        rightpos = right_min + minpos
        lefty = np.power((tmpy[:minpos] - up), 2)
        left_min = np.argmin(lefty)
        leftpos = left_min
        le = x[minpos] - x[leftpos]
        re = x[rightpos] - x[minpos]

        if band:
            plt.axvspan(x[leftpos], x[rightpos], facecolor='g', alpha=0.5)

        if text:
            plt.figtext(0.5, 0.5,
                        '%s = %7.3e ( -%7.3e , +%7.3e)' % (vname, x[minpos], le, re),
                        ha='center')
    except ValueError:
        warnings.warn(RuntimeWarning('band and text is requested but'
                                     ' the bound is too narrow.'))

    return x, y, s


def draw_contour(self, x, y, bins=20, bound=2, args=None, show_sigma=False):
    from matplotlib import pyplot as plt
    vx, vy, vz = self.contour(x, y, bins, bound, args, subtract_min=True)

    v = [self.errordef * (i ** 2) for i in range(1, 4)]

    CS = plt.contour(vx, vy, vz, v, colors=['b', 'k', 'r'])
    if not show_sigma:
        plt.clabel(CS, v)
    else:
        tmp = dict((vv, r'%i $\sigma$' % (i + 1)) for i, vv in enumerate(v))
        plt.clabel(CS, v, fmt=tmp, fontsize=16)
    plt.xlabel(x)
    plt.ylabel(y)
    plt.axhline(self.values[y], color='k', ls='--')
    plt.axvline(self.values[x], color='k', ls='--')
    plt.grid(True)
    return vx, vy, vz


def mncontour_grid(self, x, y, numpoints=20, nsigma=2, sigma_res=4, bins=100, edges=False):
    import numpy as np
    from matplotlib import mlab
    dfcn = []
    xps = []
    yps = []
    sigmas = np.linspace(0.1, nsigma + 0.5, sigma_res * nsigma)
    for i, this_sig in enumerate(sigmas):
        xminos, yminos, pts = self.mncontour(x, y, numpoints=numpoints,
                                             sigma=this_sig)
        if len(pts) == 0:  # pragma: no cover
            warnings.warn(RuntimeWarning('Fail mncontour for %s, %s, sigma=%f' % (x, y, this_sig)))
            continue

        xp, yp = zip(*pts)
        xps.append(xp)
        yps.append(yp)
        dfcn.append([this_sig] * len(pts))

    # add the minimum in
    xps.append([self.values[x]])
    yps.append([self.values[y]])
    dfcn.append([0])

    # making grid data x ,y z
    fx, fy, fz = np.hstack(xps), np.hstack(yps), np.hstack(dfcn)

    maxx, minx = max(fx), min(fx)
    maxy, miny = max(fy), min(fy)

    # extend bound a bit
    maxx += 0.05 * (maxx - minx)
    minx -= 0.05 * (maxx - minx)
    maxy += 0.05 * (maxy - miny)
    miny -= 0.05 * (maxy - miny)

    # need constant spacing for linear
    xgrid = np.linspace(minx, maxx, bins)
    ygrid = np.linspace(miny, maxy, bins)
    xstep = (maxx - minx) / (1.0 * bins)
    ystep = (maxy - miny) / (1.0 * bins)
    # xgrid = np.arange(minx, maxx+xstep/2, xstep) # over cover
    # ygrid = np.arange(miny, maxy+ystep/2, ystep)

    with warnings.catch_warnings():
        # suppress FutureWarning from matplotlib/tri/triangulation.py
        warnings.filterwarnings("ignore", category=FutureWarning)
        g = mlab.griddata(fx, fy, fz, xgrid, ygrid, interp='linear')

    # return grid edges instead of mid point (for pcolor)
    if edges:  # pragma: no cover 
        xgrid -= xstep / 2.
        ygrid -= ystep / 2.
        np.resize(xgrid, len(xgrid) + 1)
        np.resize(ygrid, len(ygrid) + 1)
        xgrid[-1] = xgrid[-2] + xstep / 2.
        ygrid[-1] = ygrid[-2] + ystep / 2.

    return xgrid, ygrid, g, (xps, yps, dfcn)


def draw_mncontour(self, x, y, bins=100, nsigma=2, numpoints=20, sig_res=4):
    from matplotlib import pyplot as plt

    xgrid, ygrid, g, r = mncontour_grid(self, x, y, numpoints, nsigma, sig_res, bins)
    # g[g.mask] = nsigma+1 #remove the mask

    CS = plt.contour(xgrid, ygrid, g, range(1, nsigma + 1))
    plt.clabel(CS, inline=1, fontsize=10)
    plt.xlabel(x)
    plt.ylabel(y)
    return xgrid, ygrid, g, CS



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from iminuit import Minuit
from nose.tools import (assert_almost_equal,
                        assert_less
                        )

import matplotlib as mpl

mpl.use('Agg')


def f1(x, y):
    return (1 - x) ** 2 + 100 * (y - 1) ** 2


def test_mnprofile():
    m = Minuit(f1, x=0, y=0, pedantic=False, print_level=1)
    m.tol = 1e-4
    m.migrad()
    assert_less(m.fval, 1e-6)
    assert_almost_equal(m.values['x'], 1., places=3)
    assert_almost_equal(m.values['y'], 1., places=3)
    m.minos('x')
    m.draw_mnprofile('x')


def test_mncontour():
    m = Minuit(f1, x=0, y=0, pedantic=False, print_level=1)
    m.tol = 1e-4
    m.migrad()
    assert_less(m.fval, 1e-6)
    assert_almost_equal(m.values['x'], 1., places=3)
    assert_almost_equal(m.values['y'], 1., places=3)
    m.minos()
    m.draw_mncontour('x', 'y')


def test_drawcontour():
    m = Minuit(f1, x=0, y=0, pedantic=False, print_level=1)
    m.tol = 1e-4
    m.migrad()
    assert_less(m.fval, 1e-6)
    assert_almost_equal(m.values['x'], 1., places=3)
    assert_almost_equal(m.values['y'], 1., places=3)
    m.minos()
    m.draw_contour('x', 'y')


def test_drawcontour_show_sigma():
    m = Minuit(f1, x=0, y=0, pedantic=False, print_level=1)
    m.tol = 1e-4
    m.migrad()
    assert_less(m.fval, 1e-6)
    assert_almost_equal(m.values['x'], 1., places=3)
    assert_almost_equal(m.values['y'], 1., places=3)
    m.minos()
    m.draw_contour('x', 'y', show_sigma=True)



"""Python 2 / 3 compatibility layer.
"""
import sys

py_ver = sys.version_info
PY2 = False
PY3 = False
if py_ver[0] == 2:
    PY2 = True
else:  # just in case PY4
    PY3 = True

ARRAY_DOUBLE_TYPECODE = 'd' if PY2 else u'd'


def is_string(s):
    try:  # Python 2
        return isinstance(s, basestring)
    except NameError:  # Python 3
        return isinstance(s, str)



"""IMinuit utility functions and classes.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import re
from .py23_compat import is_string

__all__ = [
    'describe',
    'Struct',
    'fitarg_rename',
    'true_param',
    'param_name',
    'extract_iv',
    'extract_limit',
    'extract_error',
    'extract_fix',
    'remove_var',
    'arguments_from_docstring',
]


class Struct:
    """A Struct is a Python dict with tab completion.

    Example:

    >>> s = Struct(a=42)
    >>> s['a']
    42
    >>> s.a
    42
    """
    def __init__(self, **kwds):
        self.__dict__.update(kwds)

    def __str__(self):
        return self.__dict__.__str__()

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, s):
        return self.__dict__[s]


def arguments_from_docstring(doc):
    """Parse first line of docstring for argument name.

    Docstring should be of the form ``min(iterable[, key=func])``.

    It can also parse cython docstring of the form
    ``Minuit.migrad(self[, int ncall_me =10000, resume=True, int nsplit=1])``
    """

    if doc is None:
        raise RuntimeError('__doc__ is None')

    doc = doc.lstrip()

    # care only the firstline
    # docstring can be long
    line = doc.split('\n', 1)[0]  # get the firstline
    if line.startswith("('...',)"):
        line = doc.split('\n', 2)[1]  # get the second line
    p = re.compile(r'^[\w|\s.]+\(([^)]*)\).*')
    # 'min(iterable[, key=func])\n' -> 'iterable[, key=func]'
    sig = p.search(line)
    if sig is None:
        return []
    # iterable[, key=func]' -> ['iterable[' ,' key=func]']
    sig = sig.groups()[0].split(',')
    ret = []
    for s in sig:
        # get the last one after all space after =
        # ex: int x= True
        tmp = s.split('=')[0].split()[-1]
        # clean up non _+alphanum character
        tmp = ''.join([x for x in tmp if x.isalnum() or x == '_'])
        ret.append(tmp)
        # re.compile(r'[\s|\[]*(\w+)(?:\s*=\s*.*)')
        # ret += self.docstring_kwd_re.findall(s)
    ret = list(filter(lambda x: x != '', ret))

    if len(ret) == 0:
        raise RuntimeError('Your doc is unparsable\n' + doc)

    return ret


def fc_or_c(f):
    if hasattr(f, 'func_code'):
        return f.func_code
    else:
        return f.__code__


def arguments_from_funccode(f):
    """Check f.funccode for arguments
    """
    fc = fc_or_c(f)
    vnames = fc.co_varnames
    # bound method and fake function will be None
    if is_bound(f):
        # bound method dock off self
        return list(vnames[1:fc.co_argcount])
    else:
        # unbound and fakefunc
        return list(vnames[:fc.co_argcount])


def arguments_from_call_funccode(f):
    """Check f.__call__.func_code for arguments
    """
    fc = fc_or_c(f.__call__)
    argcount = fc.co_argcount
    return list(fc.co_varnames[1:argcount])


def is_bound(f):
    """Test whether ``f`` is a bound function.
    """
    return getattr(f, '__self__', None) is not None


def dock_if_bound(f, v):
    """Dock off ``self`` if a bound function is passed.
    """
    return v[1:] if is_bound(f) else v


def better_arg_spec(f, verbose=False):
    """Extract function signature.

    ..seealso::

        :ref:`function-sig-label`
    """
    # using funccode
    try:
        return arguments_from_funccode(f)
    except Exception as e:
        if verbose:
            print(e)  # TODO: this might not be such a good idea.
            print("Extracting arguments from f.func_code/__code__ fails")

    # using __call__ funccode
    try:
        return arguments_from_call_funccode(f)
    except Exception as e:
        if verbose:
            print(e)  # TODO: this might not be such a good idea.
            print("Extracting arguments from f.__call__.func_code/__code__ fails")

    # try:
    #     return list(inspect.getargspec(f.__call__)[0][1:])
    # except Exception as e:
    #     if verbose:
    #         print(e)
    #         print("inspect.getargspec(f)[0] fails")

    # try: 
    #     return list(inspect.getargspec(f)[0])
    # except Exception as e:
    #     if verbose:
    #         print(e)
    #         print("inspect.getargspec(f)[0] fails")

    # now we are parsing __call__.__doc__
    # we assume that __call__.__doc__ doesn't have self
    # this is what cython gives
    try:
        t = arguments_from_docstring(f.__call__.__doc__)
        if t[0] == 'self':
            t = t[1:]
        return t
    except Exception as e:
        if verbose:
            print(e)
            print("fail parsing __call__.__doc__")

    # how about just __doc__
    try:
        t = arguments_from_docstring(f.__doc__)
        if t[0] == 'self':
            t = t[1:]
        return t
    except Exception as e:
        if verbose:
            print(e)
            print("fail parsing __doc__")

    raise TypeError("Unable to obtain function signature")


def describe(f, verbose=False):
    """Try to extract the function argument names.

    .. seealso::

        :ref:`function-sig-label`
    """
    return better_arg_spec(f, verbose)


def fitarg_rename(fitarg, ren):
    """Rename variable names in ``fitarg`` with rename function.

    ::

        #simple renaming
        fitarg_rename({'x':1, 'limit_x':1, 'fix_x':1, 'error_x':1},
            lambda pname: 'y' if pname=='x' else pname)
        #{'y':1, 'limit_y':1, 'fix_y':1, 'error_y':1},

        #prefixing
        figarg_rename({'x':1, 'limit_x':1, 'fix_x':1, 'error_x':1},
            lambda pname: 'prefix_'+pname)
        #{'prefix_x':1, 'limit_prefix_x':1, 'fix_prefix_x':1, 'error_prefix_x':1}

    """
    tmp = ren
    if is_string(ren):
        ren = lambda x: tmp + '_' + x
    ret = {}
    prefix = ['limit_', 'fix_', 'error_', ]
    for k, v in fitarg.items():
        vn = k
        pf = ''
        for p in prefix:
            if k.startswith(p):
                vn = k[len(p):]
                pf = p
        newvn = pf + ren(vn)
        ret[newvn] = v
    return ret


def true_param(p):
    """Check if ``p`` is a parameter name, not a limit/error/fix attributes.
    """
    return (not p.startswith('limit_') and
            not p.startswith('error_') and
            not p.startswith('fix_'))


def param_name(p):
    """Extract parameter name from attributes.

    Examples:

    - ``fix_x`` -> ``x``
    - ``error_x`` -> ``x``
    - ``limit_x`` -> ``x``
    """
    prefix = ['limit_', 'error_', 'fix_']
    for prf in prefix:
        if p.startswith(prf):
            return p[len(prf):]
    return p


def extract_iv(b):
    """Extract initial value from fitargs dictionary."""
    return dict((k, v) for k, v in b.items() if true_param(k))


def extract_limit(b):
    """Extract limit from fitargs dictionary."""
    return dict((k, v) for k, v in b.items() if k.startswith('limit_'))


def extract_error(b):
    """Extract error from fitargs dictionary."""
    return dict((k, v) for k, v in b.items() if k.startswith('error_'))


def extract_fix(b):
    """extract fix attribute from fitargs dictionary"""
    return dict((k, v) for k, v in b.items() if k.startswith('fix_'))


def remove_var(b, exclude):
    """Exclude variable in exclude list from b."""
    return dict((k, v) for k, v in b.items() if param_name(k) not in exclude)


def make_func_code(params=None):
    """Make a func_code object to fake function signature.

    You can make a funccode from describable object by::

        make_func_code(describe(f))
    """
    return Struct(co_varnames=params, co_argcount=len(params))



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from nose.tools import assert_equal
from iminuit.util import describe
from math import ldexp
import pyximport;

pyximport.install()
from . import cyfunc


# simple function
def f(x, y):
    pass


def test_simple():
    assert_equal(describe(f, True), ['x', 'y'])


# test bound method
class A:
    def f(self, x, y):
        pass


def test_bound():
    a = A()
    assert_equal(describe(a.f, True), ['x', 'y'])


# unbound method
def test_unbound():
    assert_equal(describe(A.f, True), ['self', 'x', 'y'])


# faking func code
class Func_Code:
    def __init__(self, varname):
        self.co_varnames = varname
        self.co_argcount = len(varname)


# test __call__
class Func1:
    def __init__(self):
        pass

    def __call__(self, x, y):
        return (x - 2.) ** 2 + (y - 5.) ** 2 + 10


def test_call():
    f1 = Func1()
    assert_equal(describe(f1, True), ['x', 'y'])


# fake func
class Func2:
    def __init__(self):
        self.func_code = Func_Code(['x', 'y'])

    def __call__(self, *arg):
        return (arg[0] - 2.) ** 2 + (arg[1] - 5.) ** 2 + 10


def test_fakefunc():
    f2 = Func2()
    assert_equal(describe(f2, True), ['x', 'y'])


# builtin (parsing doc)
def test_builtin():
    assert_equal(describe(ldexp, True), ['x', 'i'])


def test_cython_embedsig():
    assert_equal(describe(cyfunc.f, True), ['a', 'b'])


def test_cython_class_method():
    cc = cyfunc.CyCallable()
    assert_equal(describe(cc.test, True), ['c', 'd'])
